#!/usr/bin/env python

# context browser monitor
# Manuel Amador (Rudd-O)
# part of Smart DJ
# written on 2005-11-28

from ControllableThread import ControllableThread
import os
from songanalysiscommon import say_sorry
from amarokplus import NotAnalyzedYet,UnAnalyzable,DcopAmarokFailed
from Observable import Observer
import logging
from threading import Thread
from songanalysiscommon import say_sorry,n_logfile
import time
import urllib

class DelayedUpdate(Thread):
	def __init__(self,contextbrowser):
		Thread.__init__(self)
		self.contextbrowser = contextbrowser
		self.do_not_update = False
		self.setDaemon(True)
	def run(self):
		time.sleep(2)
		if self.do_not_update is False: self.contextbrowser.update = True
	def go_away(self): self.do_not_update = True


class ContextBrowserUpdater(ControllableThread,Observer):

	first = """
		function generateSongNode(url,similarity,title,tempo) {
			var row = document.createElement('tr');
			row.setAttribute('class','box-row');
			
			var cell = document.createElement('td');
			row.setAttribute('class','song');
			
			var ahref = document.createElement('a');
			ahref.setAttribute('href',url);
			
			var songspan = document.createElement('span');
			songspan.setAttribute('class','album-song-title');
			
			var songtext = document.createTextNode(title);
			songspan.appendChild(songtext);
			
			ahref.appendChild(songspan);
			
			var cell2 = document.createElement('td');
			cell2.setAttribute('class','sbtext');
			cell2.setAttribute('width',1);
			var tempotext = document.createTextNode(""+tempo+" BPM");
			var temponobr = document.createElement('nobr');
			temponobr.appendChild(tempotext);
			cell2.appendChild(temponobr);

			var cell3 = document.createElement('td');
			cell3.setAttribute('class','sbtext');
			cell3.setAttribute('width',1);
			
			var snspan = document.createElement('span');
//			snspan.setAttribute('class','album-song-trackno');
			
			var sntext = document.createTextNode(similarity);
			snspan.appendChild(sntext);
			
			var snnobr = document.createElement('nobr');
			snnobr.appendChild(snspan);
			
			cell3.appendChild(snnobr);

// 			cell.appendChild(div);
 			cell.appendChild(ahref);
 			row.appendChild(cell);
			row.appendChild(cell2);
			row.appendChild(cell3);
			return row;
		}
		
		function generateSimpleBox(text,url) {
			var row = document.createElement('tr');
			row.setAttribute('class','box-row');
			
			var cell = document.createElement('td');
			row.setAttribute('class','song');
			
//			var ahref = document.createElement('a');
//			ahref.setAttribute('href',url);
			
			var songspan = document.createElement('p');
//			songspan.setAttribute('class','album-song-title');
			
			var songtext = document.createTextNode(text);
			songspan.appendChild(songtext);
			
//			ahref.appendChild(songspan);
 			
//			cell.appendChild(document.createElement('br'));
			cell.appendChild(songspan);
//			cell.appendChild(document.createElement('br'));
 			row.appendChild(cell);
			return row;
		}
			
		try {
			var refBox = document.getElementById('current_box');
			if (refBox != null) {
			
				var testAlreadyAdded = document.getElementById('similar_box');
				if (testAlreadyAdded != null) {
					testAlreadyAdded.parentNode.removeChild(testAlreadyAdded);
				}
				
				
				var songName = document.getElementById('current_box-header-songname');
				var songSep = document.getElementById('current_box-header-separator');
				var songArtist = document.getElementById('current_box-header-artist');
				var refBoxTable = document.getElementById('current_box-table');
		
				var similarBox = document.createElement('div');
				similarBox.setAttribute('id','similar_box');
				similarBox.setAttribute('class','box');
		
				var similarBoxHeader = document.createElement('div');
				similarBoxHeader.setAttribute('id','similar_box-header');
				similarBoxHeader.setAttribute('onclick','toggleBlock("T_ST"); window.location.href="togglebox:st";');
				similarBoxHeader.setAttribute('style','cursor: pointer;');
				similarBoxHeader.setAttribute('class','box-header');
		
				var similarBoxHeaderTitle = document.createElement('span');
				similarBoxHeaderTitle.setAttribute('class','box-header-title');
				similarBoxHeaderTitle.appendChild(document.createTextNode('Similar Tracks according to Smart DJ'));
				similarBoxHeader.appendChild(similarBoxHeaderTitle);
					
				similarBoxHeader.appendChild(document.createElement('br'));
				
				var similarBoxHeaderSubtitle = document.createElement('span');
				similarBoxHeaderSubtitle.setAttribute('class','box-header-title');
				similarBoxHeaderSubtitle.setAttribute('style','font-weight: normal;');
				similarBoxHeaderSubtitle.appendChild(document.createTextNode('%s'));
				similarBoxHeader.appendChild(similarBoxHeaderSubtitle);
				
		//		similarBoxHeader.appendChild(songName.cloneNode(true));
		//		similarBoxHeader.appendChild(document.createTextNode(' '));
		//		similarBoxHeader.appendChild(songSep.cloneNode(true));
		//		similarBoxHeader.appendChild(document.createTextNode(' '));
		//		similarBoxHeader.appendChild(songArtist.cloneNode(true));
		
				var similarBoxTable = refBoxTable.cloneNode(true);
				similarBoxTable.setAttribute('id','T_ST');
				similarBoxTable.setAttribute('class','box-body');
		
				var similarBoxTBody = similarBoxTable.firstChild
		
				while (similarBoxTBody.hasChildNodes()) {
					similarBoxTBody.removeChild(similarBoxTBody.firstChild);
				}
				
				%s
				
				similarBox.appendChild(similarBoxHeader);
				similarBox.appendChild(similarBoxTable);
		
				refBox.parentNode.insertBefore(similarBox,refBox.nextSibling);
			}
		} catch (e) {
			alert('An error occurred updating the context browser Current tab\\nThe following message box will contain more information.\\nPlease report this error to the Smart DJ developers.');
			alert(e);
		}
"""

	second = """
		var node =
		generateSongNode("%(url)s","%(number)s","%(title)s","%(bpm)s");
		similarBoxTBody.appendChild(node);
		"""

	third = """
		var node =
		generateSimpleBox("%s","%s");
		similarBoxTBody.appendChild(node);
		"""
	logger = logging.getLogger("ContextBrowserUpdater")

	def __init__(self,proxyObject,analysisProcess,configObject):
		ControllableThread.__init__(self)
		self.configObject = configObject
		self.analysisProcess = analysisProcess
		self.amarok = proxyObject
		self.observe(self.amarok)
		self.observe(self.amarok.contextBrowser)
		self.amarok_available_yet = False
		self.observe(self.analysisProcess)
		self.update = False
		self.flag_string = "<div id='current_box' class='box'>"
		self.pending_song = None

	def notify(self,object,eventName,*args):
		if eventName == "amaroKExit":
			self.kill(asynchronous=True) # we kill ourselves
		if eventName == "amaroKAvailable":
			self.amarok_available_yet = True
		if eventName == "contentsChanged" and object == self.amarok.contextBrowser:
			if hasattr(self,"delayed_update"):
				self.delayed_update.go_away()
				delattr(self,"delayed_update")
			self.delayed_update = DelayedUpdate(self)
			self.delayed_update.start()
		if eventName == "trackAnalyzed" and self.pending_song == args[0]:
			self.pending_song = None
			self.update = True
	
	def run(self):
# 		while not self.__finish: time.sleep(1)
		try: self._run()
		except DcopAmarokFailed,e:
			self.logger.debug("Aborting because amaroK is unexpectedly gone: %s",e.output)
		except Exception,e:
			self.logger.exception("Something wrong in context browser monitor, process stopped %s %s",e,e.__class__)
			say_sorry("Something went wrong in the context browser monitor loop.  This is not your fault, but a bug in the Smart DJ plug-in.  Please report the contents of the file %s to the Smart DJ plug-in developers." %os.path.expanduser(n_logfile))
		except:
			self.logger.exception("Something wrong in context browser monitor, process stopped")
			say_sorry("Something went wrong in the context browser monitor loop.  This is not your fault, but a bug in the Smart DJ plug-in.  Please report the contents of the file %s to the Smart DJ plug-in developers." %os.path.expanduser(n_logfile))
	
	def _run(self):
		
		self.logger.debug("Context browser monitor starting")
		while not self.amarok_available_yet and not self.killed(): time.sleep(1)
		self.logger.debug("Context browser monitor started")
		
		self.perform_js_injection() # do once when we start
		while not self.killed():
			if self.update is False:
				time.sleep(1)
				continue
			self.update = False # we do this so we can rerun it on the next iteration ASAP
			self.perform_js_injection()
		
		self.logger.debug("Context browser monitor done")

	def perform_js_injection(self):
		if self.killed(): return
		htmlcontents = file(self.amarok.contextBrowser.getFilePath(),"rb").read(-1)
		if not self.flag_string in htmlcontents:
			self.logger.debug( "file is not current context browser, ignoring" )
			return
		
		if self.killed(): return
		self.logger.debug("file is current context browser, going forward with machiavellic plan")
		start_file = self.amarok.player.path()
		
		if not start_file: return
		self.logger.debug("Using %s as start file",start_file)
		if self.killed(): return
		
		c = self.configObject["super_dynamic_mode"]
		if c["enabled"]:
			spectrum_weight,tempo_weight = (c["spectrum_weight"],c["tempo_weight"])
		else: spectrum_weight,tempo_weight = (1.0,1.0)
		weighting_text = "Criteria: %d%% tempo, %d%% music style"%(
				round(tempo_weight/2*100),round(spectrum_weight/2*100))
		
		def escape_js(string):
			return string.replace("\\","\\\\").replace("\"","\\\"")
		def makeurl(path):
			return "file://" + urllib.pathname2url(path)
		
		try:
			results = self.amarok.collection.getSimilarSongs(start_file,spectrum_weight,tempo_weight,11)
		except NotAnalyzedYet,e:
			self.analysisProcess.addToPriorityQueue(start_file)
			self.pending_song = start_file # remember it for the next run
			if self.killed(): return
			start_song = self.amarok.collection.getSong(start_file)
			if self.killed(): return
			self.amarok.playlist.shortStatusMessage(
				"Smart DJ is analyzing %s to find similar songs"%start_song.getLongName(True))
			self.logger.debug("%s will be deferred because it was not analyzed yet",start_file)
			
			if self.killed(): return
			totalcommand = self.first%(escape_js(weighting_text),self.third%(escape_js("%s is being analyzed to find similar tracks..."%start_song.getShortName()),makeurl(start_file)))
			if self.killed(): return
			self.logger.debug("Submitting JavaScript") #: %s",totalcommand)
			self.amarok.contextBrowser.evalJS(totalcommand)
			
			return
		except UnAnalyzable,e:
			self.logger.debug("%s will be disregarded because its analysis failed",start_file)
			
			if self.killed(): return
			start_song = self.amarok.collection.getSong(start_file)
			if self.killed(): return
			totalcommand = self.first%(escape_js(weighting_text),self.third%(escape_js("%s was analyzed but the analysis failed, probably because its audio format is not supported by songanalysis."%start_song.getShortName()),makeurl(start_file)))
			if self.killed(): return
			self.logger.debug("Submitting JavaScript") #: %s",totalcommand)
			self.amarok.contextBrowser.evalJS(totalcommand)
			
			return
		if len(results) < 2:
			self.logger.debug( "No songs similar to %s found" ,start_file)
			if self.killed(): return
			start_song = self.amarok.collection.getSong(start_file)
			if self.killed(): return
			totalcommand = self.first%(escape_js(weighting_text),self.third%(escape_js("No tracks similar to %s were found."%start_song.getShortName()),makeurl(start_file)))
			if self.killed(): return
			self.logger.debug("Submitting JavaScript") #: %s",totalcommand)
			self.amarok.contextBrowser.evalJS(totalcommand)
			return
		
		if self.killed(): return
		
		topcompat = max([ r[1] for r in results ])
		if results[0][0] == start_file: results = results[1:] # we peel off the first result if it's the same song
		paths = [ r[0] for r in results ]
		titles = []
		
		for path in paths:
			if self.killed(): return
			song = self.amarok.collection.getSong(path)
			title = song.getShortName() # i should probably escape < > and & here as well
			titles.append(escape_js(title))
		assert ( len(results) == len(titles) )
		if self.killed(): return
		
		urls = [ makeurl(p) for p in paths ]
		compats = [ r[1] for r in results ]
		percents = [ "%d"%(p*100.0/topcompat) +"%" for p in compats ]
		tempos = [ int(round(r[2])) for r in results ]
		parameters_list = [ { "url":urls[r] , "number": percents[r] , "title": titles[r] , "bpm" : tempos[r] } for r in range(len(results)) ]
		totalcommand = self.first%(weighting_text,"\n\n".join([ self.second%p for p in parameters_list ]))
		
		if self.killed(): return
		self.logger.debug("Submitting JavaScript") #: %s",totalcommand)
		self.amarok.contextBrowser.evalJS(totalcommand)

